'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

var __assign = function() {
    __assign = Object.assign || function __assign(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p)) t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

function __values(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
}

function __read(o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
}

function __spreadArray(to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
}

var DIRECTIVE_PREFIX = 'l-';
var COMPONENT_FLAG = 'component';
var FOR_TEMPLATE_FLAG = '__for_template';
var MODEL_REGISTERED_FLAG = '__model_registered';
var DIRECTIVE_SHORTHANDS;

(function (DIRECTIVE_SHORTHANDS) {
  DIRECTIVE_SHORTHANDS["@"] = "on";
  DIRECTIVE_SHORTHANDS[":"] = "bind";
})(DIRECTIVE_SHORTHANDS || (DIRECTIVE_SHORTHANDS = {}));

var ASTNodeType;

(function (ASTNodeType) {
  ASTNodeType[ASTNodeType["NULL"] = -1] = "NULL";
  ASTNodeType[ASTNodeType["STATIC"] = 0] = "STATIC";
  ASTNodeType[ASTNodeType["DYNAMIC"] = 1] = "DYNAMIC";
})(ASTNodeType || (ASTNodeType = {}));

var error = function error(err, expression, el) {
  var message = "Lucia Error: \"" + err + "\"";

  if (expression) {
    message += "\n\nExpression: \"" + expression + "\"";
  }

  if (el) {
    message += "\nElement:";
  }

  console.warn(message, el);
};

var computeExpression = function computeExpression(expression, el, returnable, refs) {
  if (returnable === void 0) {
    returnable = true;
  }

  if (refs === void 0) {
    refs = {};
  }

  var formattedExpression = "" + (returnable ? "return " + expression : expression);
  var customGlobalSpecialProperties = window.LuciaSpecialProperties || {};

  var specialPropertiesNames = __spreadArray(['$state', '$el', '$emit', '$event', '$refs'], __read(Object.keys(customGlobalSpecialProperties)));

  var computeFunction = new Function("return function(" + specialPropertiesNames.join(',') + "){with($state){" + formattedExpression + "}}")();

  var emit = function emit(name, options, dispatchGlobal) {
    if (dispatchGlobal === void 0) {
      dispatchGlobal = true;
    }

    var event = new CustomEvent(name, options);
    var target = dispatchGlobal ? window : el || window;
    target.dispatchEvent(event);
  };

  return function (state, event) {
    try {
      var value = state[expression];

      if (value) {
        return typeof value === 'function' ? value.bind(state)() : value;
      } else {
        return computeFunction.apply(void 0, __spreadArray([state, el, emit, event, refs], __read(Object.values(customGlobalSpecialProperties))));
      }
    } catch (err) {
      error(err, expression, el);
    }
  };
};

var rawDirectiveSplitRE = function rawDirectiveSplitRE() {
  return /:|\./gim;
};
var eventDirectivePrefixRE = function eventDirectivePrefixRE() {
  return /on|@/gim;
};
var parenthesisWrapReplaceRE = function parenthesisWrapReplaceRE() {
  return /\(|\)/gim;
};
var hasDirectiveRE = function hasDirectiveRE() {
  return new RegExp("(" + DIRECTIVE_PREFIX + "|" + Object.keys(DIRECTIVE_SHORTHANDS).join('|') + ")\\w+", 'gim');
};
var expressionPropRE = function expressionPropRE(prop) {
  return new RegExp("\\b" + prop + "\\b", 'gim');
};

var removeDupesFromArray = function removeDupesFromArray(array) {
  return __spreadArray([], __read(new Set(array)));
};

var isListRenderScope = function isListRenderScope(el) {
  return el.hasAttribute(DIRECTIVE_PREFIX + "for");
};
var isUnderListRenderScope = function isUnderListRenderScope(el) {
  return !!el.parentElement && el.parentElement.hasAttribute(DIRECTIVE_PREFIX + "for");
};
var createASTNode = function createASTNode(el, state) {
  var _a = __read(collectAndInitDirectives(el, state), 2),
      directives = _a[0],
      deps = _a[1];

  var hasDirectives = Object.keys(directives).length > 0;
  var hasDepInDirectives = Object.values(directives).some(function (_a) {
    var value = _a.value;
    return Object.keys(state).some(function (prop) {
      return expressionPropRE(prop).test(value);
    });
  });
  var type = hasDepInDirectives ? ASTNodeType.DYNAMIC : ASTNodeType.STATIC;
  var node = {
    el: el,
    deps: deps,
    directives: directives,
    type: type
  };
  return hasDirectives ? node : undefined;
};
var collectRefs = function collectRefs() {
  var refDirective = DIRECTIVE_PREFIX + "ref";
  var refElements = document.querySelectorAll("[" + refDirective + "]");
  var refs = {};
  var _a2 = refElements;

  var _f = function _f(refElement) {
    var name = refElement.getAttribute(refDirective);

    if (name) {
      refs[name] = refElement;
    }
  };

  for (var _i = 0; _i < _a2.length; _i++) {
    _f(_a2[_i]);
  }
  return refs;
};
var collectAndInitDirectives = function collectAndInitDirectives(el, state) {
  var e_1, _a;

  if (state === void 0) {
    state = {};
  }

  var directives = {};
  var refs = collectRefs();
  var nodeDeps = [];

  var _loop_1 = function _loop_1(name_1, value) {
    var isStateDirective = name_1 === DIRECTIVE_PREFIX + "state";
    var hasDirectivePrefix = name_1.startsWith(DIRECTIVE_PREFIX);
    var hasDirectiveShorthandPrefix = Object.keys(DIRECTIVE_SHORTHANDS).some(function (shorthand) {
      return name_1.startsWith(shorthand);
    });

    if (isStateDirective || !(hasDirectivePrefix || hasDirectiveShorthandPrefix)) {
      return "continue";
    }

    var depsInFunctions = [];
    var propsInState = Object.keys(state);
    var returnable = true;
    var deps = propsInState.filter(function (prop) {
      var hasDep = expressionPropRE(prop).test(String(value));

      if (hasDep && typeof state[prop] === 'function') {
        var depsInFunction = propsInState.filter(function (p) {
          return expressionPropRE(p).test(String(state[prop]));
        });
        depsInFunctions.push.apply(depsInFunctions, __spreadArray([], __read(depsInFunction)));
      }

      return hasDep;
    });

    if (eventDirectivePrefixRE().test(name_1)) {
      returnable = false;
    }

    if (name_1.includes('for') && el[FOR_TEMPLATE_FLAG] === undefined) {
      el[FOR_TEMPLATE_FLAG] = String(el.innerHTML).trim();
      returnable = false;
    }

    var uniqueCompiledDeps = removeDupesFromArray(__spreadArray(__spreadArray([], __read(deps)), __read(depsInFunctions)));
    nodeDeps.push.apply(nodeDeps, __spreadArray([], __read(uniqueCompiledDeps)));
    var directiveData = {
      compute: computeExpression(value, el, returnable, refs),
      deps: uniqueCompiledDeps,
      value: value
    };
    var directiveName = hasDirectivePrefix ? name_1.slice(DIRECTIVE_PREFIX.length) : DIRECTIVE_SHORTHANDS[name_1[0]] + ":" + name_1.slice(1);
    directives[directiveName.toLowerCase()] = directiveData;
  };

  try {
    for (var _b = __values(el.attributes), _c = _b.next(); !_c.done; _c = _b.next()) {
      var _d = _c.value,
          name_1 = _d.name,
          value = _d.value;

      _loop_1(name_1, value);
    }
  } catch (e_1_1) {
    e_1 = {
      error: e_1_1
    };
  } finally {
    try {
      if (_c && !_c.done && (_a = _b["return"])) {
        _a.call(_b);
      }
    } finally {
      if (e_1) {
        throw e_1.error;
      }
    }
  }

  return [directives, removeDupesFromArray(nodeDeps)];
};
var flattenElementChildren = function flattenElementChildren(rootElement, isListGroup, ignoreRootElement) {
  var e_2, _a;

  if (isListGroup === void 0) {
    isListGroup = false;
  }

  if (ignoreRootElement === void 0) {
    ignoreRootElement = false;
  }

  var collection = [];
  var isList = isListRenderScope(rootElement);
  var isUnderList = isUnderListRenderScope(rootElement);

  if (!isListGroup && (isList || isUnderList)) {
    return collection;
  }

  if (!ignoreRootElement && (!isListGroup || !isList)) {
    collection.push(rootElement);
  }

  if (isListGroup || !isList && !isUnderList) {
    try {
      for (var _b = __values(rootElement.children), _c = _b.next(); !_c.done; _c = _b.next()) {
        var childElement = _c.value;

        if (childElement instanceof HTMLElement) {
          if (!isListGroup && isListRenderScope(childElement)) {
            collection.push(childElement);
          } else {
            if (childElement.hasAttribute(DIRECTIVE_PREFIX + "state")) {
              continue;
            }

            collection.push.apply(collection, __spreadArray([], __read(flattenElementChildren(childElement, isListGroup, childElement.attributes.length === 0))));
          }
        }
      }
    } catch (e_2_1) {
      e_2 = {
        error: e_2_1
      };
    } finally {
      try {
        if (_c && !_c.done && (_a = _b["return"])) {
          _a.call(_b);
        }
      } finally {
        if (e_2) {
          throw e_2.error;
        }
      }
    }
  }

  return collection;
};
var compile = function compile(el, state, ignoreRootElement) {
  if (state === void 0) {
    state = {};
  }

  if (ignoreRootElement === void 0) {
    ignoreRootElement = false;
  }

  var ast = [];
  var isListGroup = el[COMPONENT_FLAG] !== undefined && isListRenderScope(el);
  var elements = flattenElementChildren(el, isListGroup, ignoreRootElement);
  var maskDirective = DIRECTIVE_PREFIX + "mask";
  var _a3 = elements;

  var _f2 = function _f2(element) {
    if (element.hasAttribute(maskDirective)) {
      element.removeAttribute(maskDirective);
    }

    if (hasDirectiveRE().test(element.outerHTML)) {
      var newASTNode = createASTNode(element, state);

      if (newASTNode) {
        ast.push(newASTNode);
      }
    }
  };

  for (var _i2 = 0; _i2 < _a3.length; _i2++) {
    _f2(_a3[_i2]);
  }
  return ast;
};

function _typeof(obj) {
  "@babel/helpers - typeof";

  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

var formatAcceptableWhitespace = function formatAcceptableWhitespace(expression) {
  var whitespaceRE = /\s+/gim;
  return expression.replace(whitespaceRE, ' ').trim();
};
var bindDirective = function bindDirective(_a) {
  var el = _a.el,
      parts = _a.parts,
      data = _a.data,
      state = _a.state;

  switch (parts[1]) {
    case 'class':
      {
        var classes = data.compute(state);
        var removeDynamicClassesRE = new RegExp("\\b" + Object.keys(classes).join('|') + "\\b", 'gim');
        var rawClasses = el.className.replace(removeDynamicClassesRE, '');

        if (typeof classes === 'string') {
          el.className = formatAcceptableWhitespace(rawClasses + " " + classes);
        } else if (Array.isArray(classes)) {
          el.className = formatAcceptableWhitespace(rawClasses + " " + classes.join(' '));
        } else {
          var activeClasses_1 = [];

          var _a2 = Object.entries(classes);

          var _f = function _f(_a) {
            var _b = __read(_a, 2),
                className = _b[0],
                classValue = _b[1];

            if (classValue) {
              activeClasses_1.push(className);
            }
          };

          for (var _i = 0; _i < _a2.length; _i++) {
            _f(_a2[_i], _i, _a2);
          }

          if (activeClasses_1.length > 0) {
            el.className = formatAcceptableWhitespace(rawClasses + " " + activeClasses_1.join(' '));
          } else if (formatAcceptableWhitespace(rawClasses).length > 0) {
            el.className = formatAcceptableWhitespace(rawClasses);
          } else {
            el.className = '';
          }
        }

        break;
      }

    case 'style':
      {
        var styles = data.compute(state);

        if (el.hasAttribute('style')) {
          el.removeAttribute('style');
        }

        var _a3 = Object.entries(styles);

        var _f2 = function _f2(_a) {
          var _b = __read(_a, 2),
              styleName = _b[0],
              styleValue = _b[1];

          el.style[styleName] = styleValue;
        };

        for (var _i2 = 0; _i2 < _a3.length; _i2++) {
          _f2(_a3[_i2], _i2, _a3);
        }
        break;
      }

    default:
      {
        var attributes = data.compute(state);

        if (_typeof(attributes) === 'object' && attributes !== null) {
          var _a4 = Object.entries(attributes);

          var _f3 = function _f3(_a) {
            var _b = __read(_a, 2),
                name = _b[0],
                value = _b[1];

            if (value) {
              el.setAttribute(name, value);
            } else {
              el.removeAttribute(name);
            }
          };

          for (var _i3 = 0; _i3 < _a4.length; _i3++) {
            _f3(_a4[_i3], _i3, _a4);
          }
        } else if (attributes) {
          el.setAttribute(parts[1], attributes);
        } else {
          if (el.hasAttribute(parts[1])) {
            el.removeAttribute(parts[1]);
          }
        }

        break;
      }
  }
};

var lazy = function lazy(threshold, generatorFunction) {
  var generator = generatorFunction();
  return function next() {
    var start = performance.now();
    var task = null;

    do {
      task = generator.next();
    } while (performance.now() - start < threshold && !task.done);

    if (task.done) {
      return;
    }

    setTimeout(next);
  };
};

var render = function render(ast, directives, state, changedProps) {
  if (state === void 0) {
    state = {};
  }

  if (changedProps === void 0) {
    changedProps = [];
  }

  var legalDirectiveNames = Object.keys(directives);
  var LAZY_MODE_TIMEOUT = 25;
  lazy(LAZY_MODE_TIMEOUT, function () {
    var _loop_1, ast_1, ast_1_1, node, e_1_1;

    var e_1, _a;

    return __generator(this, function (_b) {
      switch (_b.label) {
        case 0:
          _loop_1 = function _loop_1(node) {
            var isStatic, nodeHasDep, _loop_2, _c, _d, _e, directiveName, directiveData, e_2_1, effectEvent;

            var e_2, _f;

            return __generator(this, function (_g) {
              switch (_g.label) {
                case 0:
                  if (node.type === ASTNodeType.NULL) {
                    return [2, "continue"];
                  }

                  isStatic = node.type === ASTNodeType.STATIC;

                  if (isStatic) {
                    node.type = ASTNodeType.NULL;
                  }

                  return [4];

                case 1:
                  _g.sent();

                  nodeHasDep = changedProps.some(function (prop) {
                    return node.deps.includes(prop);
                  });

                  if (!nodeHasDep && !isStatic) {
                    return [2, "continue"];
                  }

                  _loop_2 = function _loop_2(directiveName, directiveData) {
                    var rawDirectiveName, directiveHasDep, isMaskDirective, isStaticDirective, directiveProps;
                    return __generator(this, function (_h) {
                      switch (_h.label) {
                        case 0:
                          rawDirectiveName = directiveName.split(rawDirectiveSplitRE())[0];

                          if (!legalDirectiveNames.includes(rawDirectiveName.toUpperCase())) {
                            return [2, "continue"];
                          }

                          return [4];

                        case 1:
                          _h.sent();

                          directiveHasDep = changedProps.some(function (prop) {
                            return directiveData.deps.includes(prop);
                          });
                          isMaskDirective = directiveName === DIRECTIVE_PREFIX + "mask";
                          isStaticDirective = Object.keys(directiveData.deps).length === 0;

                          if (directiveHasDep || isStatic || isStaticDirective) {
                            directiveProps = {
                              el: node.el,
                              parts: directiveName.split(rawDirectiveSplitRE()),
                              data: directiveData,
                              node: node,
                              state: state
                            };
                            renderDirective(directiveProps, directives);

                            if (isStaticDirective || isMaskDirective) {
                              delete node.directives[directiveName];

                              if (isMaskDirective) {
                                node.el.removeAttribute(DIRECTIVE_PREFIX + "mask");
                              }
                            }
                          }

                          return [2];
                      }
                    });
                  };

                  _g.label = 2;

                case 2:
                  _g.trys.push([2, 7, 8, 9]);

                  _c = (e_2 = void 0, __values(Object.entries(node.directives))), _d = _c.next();
                  _g.label = 3;

                case 3:
                  if (!!_d.done) {
                    return [3, 6];
                  }

                  _e = __read(_d.value, 2), directiveName = _e[0], directiveData = _e[1];
                  return [5, _loop_2(directiveName, directiveData)];

                case 4:
                  _g.sent();

                  _g.label = 5;

                case 5:
                  _d = _c.next();
                  return [3, 3];

                case 6:
                  return [3, 9];

                case 7:
                  e_2_1 = _g.sent();
                  e_2 = {
                    error: e_2_1
                  };
                  return [3, 9];

                case 8:
                  try {
                    if (_d && !_d.done && (_f = _c["return"])) {
                      _f.call(_c);
                    }
                  } finally {
                    if (e_2) {
                      throw e_2.error;
                    }
                  }

                  return [7];

                case 9:
                  if (node.directives['on:effect']) {
                    effectEvent = new CustomEvent('effect');
                    node.el.dispatchEvent(effectEvent);
                  }

                  return [2];
              }
            });
          };

          _b.label = 1;

        case 1:
          _b.trys.push([1, 6, 7, 8]);

          ast_1 = __values(ast), ast_1_1 = ast_1.next();
          _b.label = 2;

        case 2:
          if (!!ast_1_1.done) {
            return [3, 5];
          }

          node = ast_1_1.value;
          return [5, _loop_1(node)];

        case 3:
          _b.sent();

          _b.label = 4;

        case 4:
          ast_1_1 = ast_1.next();
          return [3, 2];

        case 5:
          return [3, 8];

        case 6:
          e_1_1 = _b.sent();
          e_1 = {
            error: e_1_1
          };
          return [3, 8];

        case 7:
          try {
            if (ast_1_1 && !ast_1_1.done && (_a = ast_1["return"])) {
              _a.call(ast_1);
            }
          } finally {
            if (e_1) {
              throw e_1.error;
            }
          }

          return [7];

        case 8:
          return [2];
      }
    });
  })();
};

var adjustDeps = function adjustDeps(ast, currentDeps, node, directiveName) {
  var deps = [];
  var _a = ast;

  var _f = function _f(node) {
    deps.push.apply(deps, __spreadArray([], __read(node.deps)));
  };

  for (var _i = 0; _i < _a.length; _i++) {
    _f(_a[_i]);
  }
  var cleanedDeps = removeDupesFromArray(__spreadArray(__spreadArray([], __read(currentDeps)), __read(deps)));
  node.deps = cleanedDeps;
  node.directives[directiveName].deps = cleanedDeps;
};

var forDirective = function forDirective(_a) {
  var _b;

  var el = _a.el,
      data = _a.data,
      state = _a.state,
      node = _a.node;
  var originalAST = el[COMPONENT_FLAG];

  if (!originalAST) {
    el[COMPONENT_FLAG] = compile(el, state);
  }

  var forLoopRE = /\s+(?:in|of)\s+/gim;

  var _c = __read(data.value.split(forLoopRE), 2),
      expression = _c[0],
      target = _c[1];

  var _d = __read(expression === null || expression === void 0 ? void 0 : expression.trim().replace(parenthesisWrapReplaceRE(), '').split(','), 2),
      item = _d[0],
      index = _d[1];

  var currArray = (_b = state[target === null || target === void 0 ? void 0 : target.trim()]) !== null && _b !== void 0 ? _b : computeExpression(target === null || target === void 0 ? void 0 : target.trim(), el, true)(state);
  var template = el[FOR_TEMPLATE_FLAG];

  if (el.innerHTML.trim() === template) {
    el.innerHTML = '';
  }

  var arrayDiff = (currArray === null || currArray === void 0 ? void 0 : currArray.length) - el.children.length;

  if ((currArray === null || currArray === void 0 ? void 0 : currArray.length) === 0) {
    el.innerHTML = '';
  } else if (arrayDiff !== 0) {
    for (var i = Math.abs(arrayDiff); i > 0; --i) {
      if (arrayDiff < 0) {
        el.removeChild(el.lastChild);
      } else {
        var content = String(template);
        var isTable = /^[^\S]*?<(t(?:head|body|foot|r|d|h))/i.test(content);

        if (item) {
          content = content.replace(expressionPropRE("this\\." + item.trim()), target + "[" + (currArray.length - i) + "]");
        }

        if (index) {
          content = content.replace(expressionPropRE("this\\." + index.trim()), String(currArray.length - i));
        }

        if (isTable) {
          content = "<table>" + content + "</table>";
        }

        var fragment = document.createRange().createContextualFragment(content);
        el.appendChild(isTable ? fragment.firstElementChild.firstElementChild : fragment);
      }
    }

    el[COMPONENT_FLAG] = compile(el, state);
  }

  if (!originalAST) {
    adjustDeps(el[COMPONENT_FLAG], data.deps, node, 'for');
    el.removeAttribute(DIRECTIVE_PREFIX + "for");
  }

  var ast = arrayDiff === 0 ? originalAST : compile(el, state, true);
  render(ast || [], directives, state, node.deps);
};

var htmlDirective = function htmlDirective(_a) {
  var _b;

  var el = _a.el,
      data = _a.data,
      state = _a.state,
      node = _a.node;
  node = node;
  var marker = el[COMPONENT_FLAG];
  var ret = (_b = data.compute(state)) !== null && _b !== void 0 ? _b : data.value;

  if (ret !== el.innerHTML) {
    el.innerHTML = ret;

    if (hasDirectiveRE().test(ret)) {
      var ast = marker !== null && marker !== void 0 ? marker : compile(el, state, true);

      if (!marker) {
        adjustDeps(ast, data.deps, node, 'html');
      }

      render(ast, directives, state, data.deps);
      el[COMPONENT_FLAG] = ast;
    }
  }
};

var inputCallback = function inputCallback(el, hydratedValue, data, state) {
  if (el.type === 'checkbox') {
    el.value = String(el.checked);
  }

  var isNumber = typeof hydratedValue === 'number' && !isNaN(el.value);
  var isBoolean = typeof hydratedValue === 'boolean' && (el.value === 'true' || el.value === 'false');
  var isNullish = (hydratedValue === null || hydratedValue === undefined) && (el.value === 'null' || el.value === 'undefined');
  var payload;

  if (isNumber) {
    payload = Number(el.value);
  } else if (isBoolean) {
    payload = el.value === 'true';
  } else if (isNullish) {
    if (el.value === 'null') {
      payload = null;
    } else payload = undefined;
  } else {
    payload = String(el.value);
  }

  if (state[data.value]) {
    state[data.value] = payload;
  } else {
    payload = typeof payload === 'string' ? "'" + payload + "'" : payload;
    computeExpression("$state." + data.value + " = " + payload, el, true)(state);
  }

  return payload;
};
var modelDirective = function modelDirective(_a) {
  var _b;

  var awaitingTypecastEl = _a.el,
      parts = _a.parts,
      data = _a.data,
      state = _a.state;
  var el = awaitingTypecastEl;
  var hydratedValue = (_b = state[data.value]) !== null && _b !== void 0 ? _b : computeExpression(data.value, el, true)(state);
  var accessor = el.type === 'checkbox' ? 'checked' : 'value';

  if (el[accessor] !== String(hydratedValue)) {
    el[accessor] = hydratedValue;
  }

  if (!el[MODEL_REGISTERED_FLAG]) {
    var callback = function callback() {
      return inputCallback(el, hydratedValue, data, state);
    };

    el.addEventListener(parts[1] === 'debounce' ? 'change' : 'input', callback);
    el[MODEL_REGISTERED_FLAG] = true;
  }
};

var onDirective = function onDirective(_a) {
  var el = _a.el,
      parts = _a.parts,
      data = _a.data,
      state = _a.state;
  var options = {};
  var globalScopeEventProps = ['outside', 'global'];
  var eventProps = parts.slice(2);
  var EVENT_REGISTERED_FLAG = "__on_" + parts[1] + "_registered";

  if (el[EVENT_REGISTERED_FLAG]) {
    return;
  }

  var target = globalScopeEventProps.some(function (prop) {
    return String(eventProps).includes(prop);
  }) ? window : el;

  var handler = function handler(event) {
    if (event instanceof KeyboardEvent && /\d/gim.test(String(eventProps))) {
      var whitelistedKeycodes_1 = [];
      var _a2 = eventProps;

      var _f = function _f(eventProp) {
        if (!isNaN(eventProp)) {
          whitelistedKeycodes_1.push(Number(eventProp));
        }
      };

      for (var _i = 0; _i < _a2.length; _i++) {
        _f(_a2[_i], _i, _a2);
      }

      if (!whitelistedKeycodes_1.includes(event.keyCode)) {
        return;
      }
    }

    if (eventProps.includes('prevent')) {
      event.preventDefault();
    }

    if (eventProps.includes('stop')) {
      event.stopPropagation();
    }

    if (eventProps.includes('self')) {
      if (event.target !== el) {
        return;
      }
    }

    if (eventProps.includes('outside')) {
      if (el.contains(event.target)) {
        return;
      }

      if (el.offsetWidth < 1 && el.offsetHeight < 1) {
        return;
      }
    }

    data.compute(state, event);
  };

  options.once = eventProps.includes('once');
  options.passive = eventProps.includes('passive');
  target.addEventListener(parts[1], handler, options);
  el[EVENT_REGISTERED_FLAG] = true;
};

var showDirective = function showDirective(_a) {
  var el = _a.el,
      data = _a.data,
      state = _a.state;
  var ret = data.compute(state);

  if (ret !== el.style.display) {
    el.style.display = ret ? '' : 'none';
  }

  if (el.style.length === 0) {
    el.removeAttribute('style');
  }
};

var textDirective = function textDirective(_a) {
  var _b;

  var el = _a.el,
      data = _a.data,
      state = _a.state;
  var ret = (_b = data.compute(state)) !== null && _b !== void 0 ? _b : data.value;

  if (ret !== el.textContent) {
    el.textContent = ret;
  }
};

var customGlobalDirectives = window.LuciaDirectives || {};
var directives = __assign({
  BIND: bindDirective,
  HTML: htmlDirective,
  MODEL: modelDirective,
  SHOW: showDirective,
  ON: onDirective,
  TEXT: textDirective,
  FOR: forDirective
}, customGlobalDirectives);
var renderDirective = function renderDirective(props, directives) {
  directives[props.parts[0].toUpperCase()](props);
};

var arrayEquals = function arrayEquals(firstArray, secondArray) {
  return Array.isArray(firstArray) && Array.isArray(secondArray) && firstArray.length === secondArray.length && firstArray.every(function (value, i) {
    return value === secondArray[i];
  });
};
var reactive = function reactive(state, render, watchers) {
  if (watchers === void 0) {
    watchers = {};
  }

  var _a2 = ['Object', 'Array'];

  var _f = function _f(type) {
    return "[object " + type + "]";
  };

  var _r = [];

  for (var _i = 0; _i < _a2.length; _i++) {
    _r.push(_f(_a2[_i]));
  }

  var supportedObjectTypes = _r;
  var handler = {
    get: function get(target, key) {
      var ret = target[key];

      if (_typeof(ret) === 'object' && ret !== null) {
        var objectType = Object.prototype.toString.call(ret);

        if (supportedObjectTypes.includes(objectType)) {
          return new Proxy(ret, handler);
        } else {
          error("Data type " + objectType + " is not supported");
          return ret;
        }
      } else {
        return ret;
      }
    },
    set: function set(target, key, value) {
      var hasArrayMutationKey = !isNaN(Number(key)) || key === 'length';
      var props = hasArrayMutationKey ? [] : [key];

      if (Array.isArray(target) && hasArrayMutationKey) {
        var keys = Object.keys(state).filter(function (prop) {
          return Array.isArray(state[prop]) && arrayEquals(state[prop], target);
        });
        props.push.apply(props, __spreadArray([], __read(keys)));
      } else {
        if (!Object.is(target, state)) {
          var keys = Object.keys(state).filter(function (prop) {
            return Object.prototype.toString.call(state[prop]) === '[object Object]' && JSON.stringify(state[prop]).indexOf(key) > -1;
          });
          props.push.apply(props, __spreadArray([], __read(keys)));
        }
      }

      target[key] = value;
      render(props);

      var _a3 = Object.entries(watchers);

      var _f2 = function _f2(_a) {
        var _b = __read(_a, 2),
            prop = _b[0],
            watcher = _b[1];

        if (props.includes(prop)) {
          watcher();
        }
      };

      for (var _i2 = 0; _i2 < _a3.length; _i2++) {
        _f2(_a3[_i2]);
      }
      return true;
    }
  };
  return new Proxy(Object.seal(state), handler);
};

var Component = function () {
  function Component(state, watchers) {
    if (state === void 0) {
      state = {};
    }

    if (watchers === void 0) {
      watchers = {};
    }

    this.ast = [];
    this.state = state;
    this.directives = directives;
    this.watchers = watchers;
  }

  Component.prototype.mount = function (el) {
    var rootEl = typeof el === 'string' ? document.querySelector(el) || document.body : el;

    var newState = __assign(__assign({}, this.state), {
      $render: this.render.bind(this)
    });

    this.ast = compile(rootEl, this.state);
    this.state = reactive(newState, this.render.bind(this), this.watchers);
    this.render();
    rootEl[COMPONENT_FLAG] = this;
    var mountedEvent = new CustomEvent('mounted');
    rootEl.dispatchEvent(mountedEvent);
  };

  Component.prototype.render = function (props) {
    if (props === void 0) {
      props = Object.keys(this.state);
    }

    render(this.ast, directives, this.state, props);
  };

  return Component;
}();

var ComponentFactory = function () {
  function ComponentFactory(state) {
    this.state = state;
    this.watchers = {};
  }

  ComponentFactory.prototype.init = function (el) {
    var c = new Component(this.state(), this.watchers);
    c.mount(el);
    return c;
  };

  ComponentFactory.prototype.watch = function (name, callback) {
    this.watchers[name] = callback;
  };

  return ComponentFactory;
}();
var component = function component(state) {
  if (typeof state !== 'function') {
    error('State must be function that returns the state object.');
  }

  return new ComponentFactory(state);
};

var init = function init(element) {
  if (element === void 0) {
    element = document;
  }

  var stateDirective = DIRECTIVE_PREFIX + "state";
  var componentElements = element.querySelectorAll("[" + stateDirective + "]");

  var uninitializedComponents = __spreadArray([], __read(componentElements)).filter(function (el) {
    return el[COMPONENT_FLAG] === undefined;
  });

  var _a = uninitializedComponents;

  var _f = function _f(uninitializedComponent) {
    var stateExpression = uninitializedComponent.getAttribute(stateDirective);

    var state = function state() {
      return new Function("return " + stateExpression)() || {};
    };

    var currentComponent = component(state);
    currentComponent.init(uninitializedComponent);
  };

  for (var _i = 0; _i < _a.length; _i++) {
    _f(_a[_i]);
  }
};

exports.component = component;
exports.init = init;
